# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from annotation.models import Text, Excerpt, Segment, Workdetail
import os, io
from lxml import etree
import json
import time
from annotation.anot2_config import NO_INTERP, SORT_ORDER


namespaces = {
  'tei': 'http://www.tei-c.org/ns/1.0',
  'xi': 'http://www.w3.org/2001/XInclude',
  'nkjp': 'http://www.nkjp.pl/ns/1.0',
  'xml': 'http://www.w3.org/XML/1998/namespace'
}


'''Komenda na dodawanie nowych fragmentów i tekstów do bazy.
Domyślnie baza jest rozszerzana.
Można też usunąć wszystko z bazy i załadować na nowo przy użyciu opcji --clear-load'''

class Command(BaseCommand):
  help = 'Clears the database and/or adds new texts and excerpts to it.'
  missing_args_message = 'Please provide "texts_dir" argument'
  
  def add_arguments(self, parser):
    # Positional argument
    parser.add_argument('texts_dir', nargs='?')

    # Named (optional) arguments
    parser.add_argument('--clear-load',
      action='store_true',
      dest='clear_load',
      default=False,
      help='Clear a database and load texts from texts_dir.')
    
    parser.add_argument('--clear-db',
      action='store_true',
      dest='clear_db',
      default=False,
      help='Clear a database.')

      

  def handle(self, *args, **options):
    
    #print options
    texts_dir = unicode(options.get('texts_dir')) #None, jeśli texts_dir nie był podany argument
      
    if options['clear_load']:
      self.stdout.write('Start clear and load, texts_dir: %s' % texts_dir)
      self.__clear()
      self.__load(texts_dir)
      self.stdout.write('Successfully cleared the database and loaded new texts from directory: %s' % texts_dir)
    elif options['clear_db']:
      self.stdout.write('Start clearing the database')
      self.__clear()
      self.stdout.write('Successfully cleared the database')      
    else:
      self.stdout.write('Start adding new texts, texts_dir: %s' % texts_dir)
      self.__load(texts_dir)
      self.stdout.write('Successfully added new texts from directory: %s' % texts_dir)

  #czyści wszystkie informacje o fragmentach z bazy
  @transaction.atomic
  def __clear(self):
    self.stdout.write('Clear.')
    Text.objects.all().delete()
    Excerpt.objects.all().delete()
    Segment.objects.all().delete()
    Workdetail.objects.all().delete()
    
  #ładuje nowe teksty i fragmenty do bazy z podanego katalogu
  @transaction.atomic
  def __load(self, texts_dir):
    try:
      s = time.clock()
      self.stdout.write(u'Loading texts from {}...'.format(texts_dir))
#      text_list = os.listdir(unicode(texts_dir))
      for dirname, subdirs, files in os.walk(texts_dir):
        if ('text.xml' in files) and ('ann_morphosyntax.xml' in files):
          dir_split = unicode(dirname).rsplit('/',2)
          if dir_split[-1] == 'selection':
            text_id = dir_split[-2]
          else:
            text_id = dir_split[-1]
          self.stdout.write(u'Processing text: {}'.format(text_id))
          text_exists = Text.objects.filter(text_name = text_id).exists()
          if text_exists == False:
            with io.open(os.path.join(dirname, 'header.xml'), 'r', encoding='utf-8') as header_file:
              header_text = header_file.read()
            text_obj = Text(text_name = text_id, header = header_text, dirname = dirname)
            text_obj.save()
          else:
            text_obj = Text.objects.get(text_name = text_id) 
          self.__load_excerpts(dirname, text_obj)
      self.stdout.write("Loading time: {}".format(time.clock()-s))
    except:
      self.stdout.write(u'Failed to read text info from %s' % texts_dir)
      self.stdout.write(u'Removing text info from the database')
      raise

  #zwraca true jeśli segment podany jako argument jest ostatni
  #(lub może być ostatni - choice) w ramach zdania
  def is_last(self, xml_seg):
    ret = False
    xml_choice = xml_seg.xpath('ancestor::tei:choice', namespaces = namespaces)
    if len(xml_choice) == 0: #segment bezpośrednio w próbce
      if xml_seg.getnext() is None:
        ret = True
    elif xml_choice[0].getnext() is None: #segment gdzieś wewnątrz ostatniego choice w zdaniu
      if xml_seg.getparent() == xml_choice: #segment bezpośrednio w choice
        ret = True
      elif xml_seg.getnext() is None: #segment musi być w paren, czyli trzeba złapać ostatni
        ret = True
    return ret

  def tag_wo_ns(self, elem):
    return elem.tag.rsplit("}", 1)[-1]
    
    
  def make_interps_JSON(self, seg, xml_interps):
    try:
      os.remove('tagset_err.txt')
    except:
      pass
    interps = []
    for xml_lex in xml_interps:
      lex = {}
      lex['base'] = xml_lex.xpath('./tei:f[@name="base"]/tei:string', namespaces = namespaces)[0].text
      lex['ctag'] = xml_lex.xpath('./tei:f[@name="ctag"]/tei:symbol/@value', namespaces = namespaces)[0]
      lex['msd'] = []
      
      xml_msds = xml_lex.xpath('./tei:f[@name="msd"]//tei:symbol', namespaces = namespaces)
      for xml_msd in xml_msds:
        msd = {}
        msd['val'] = xml_msd.get('value')
        for e in msd['val'].split(":"):
          if (e != "") and (e not in SORT_ORDER):
            t_err = io.open('tagset_err.txt', 'a')
            t_err.write(u'base: {}, ctag: {}, msd: {}, kod: {}\n'.format(lex['base'], lex['ctag'], msd['val'], e))
            t_err.close()
        msd['manual'] = False
        lex['msd'].append(msd)
      lex['msd'].sort(key=lambda elem : ":".join(map(lambda e : e.replace(e, '{:02d}'.format(SORT_ORDER.index(e)+1)) if e in SORT_ORDER else "00", elem['val'].split(":"))) if len(elem['val']) > 0 else "00" )
         
      interps.append(lex)
      
      # Jeśli jedyną interpretacją jest interpuncja (ctag=interp), to ustaw ją od razu w bazie
      if (len(xml_interps) == 1) and (len(xml_msds) == 1) and (lex['ctag'] == "interp"):
        seg.base = lex['base']
        seg.ctag = "interp"
        seg.msd = ""
        seg.radio_id = "0.0"
    interps.sort(key=lambda elem : elem['base'] + ':' + elem['ctag'])
    return json.dumps(interps)
    

  def find_morph(self, seg_id, morph_index, morph_list):
    m_index = morph_index
    while (m_index < len(morph_list)) and (morph_list[m_index].get('corresp','') <> "ann_segmentation.xml#{}".format(seg_id)):
      m_index += 1
    if m_index == len(morph_list):
      return (morph_index, []) 
    else:
      return (m_index, [morph_list[m_index]])
    
    
  def make_segment(self, exc, entry_id, xml_seg, segs, choice_id, paren_id, xml_morph_seg):
    #stwórz nowy obiekt Segment
    #seg = Segment(excerpt = exc, seg_entry_id = entry_id, xml_id = segs) 
    seg = Segment(excerpt = exc, xml_id = entry_id, korbeusz = True) 
    #seg_id = xml_seg.get('{http://www.w3.org/XML/1998/namespace}id')
    corresp = xml_seg.get('corresp')
    if corresp is not None:
      seg.corresp_elem, seg.corresp_pos, seg.corresp_len = corresp.split('(')[1][:-1].split(',')
    #jeśli trzeba, ustaw pole choice_id
    if choice_id != -1:
      seg.choice_id = choice_id
    #jeśli trzeba, ustaw pole paren_id
    if paren_id != -1:
      seg.paren_id = paren_id
    #jeśli trzeba, ustaw pole 'nps'
    xml_nps = xml_seg.get('{http://www.nkjp.pl/ns/1.0}nps')
    if xml_nps is not None:
      seg.nps = True
    #jeśli trzeba, ustaw pole 'type'
    xml_type = xml_seg.get('type')
    if xml_type is not None:
      seg.seg_type = xml_type
    #jeśli ostatni w zdaniu, ustaw koniec zdania, ewentualnie koniec ab
    if self.is_last(xml_seg):
      seg.sent_end = True

    #dla elementów niepustych (mających opis w ann_morphosyntax) ustaw pola 'orth' i 'translit'
    if len(xml_morph_seg) <> 0:
      seg.translit = xml_morph_seg[0].xpath('./tei:fs/tei:f[@name="translit"]/tei:string', namespaces = namespaces)[0].text
      if xml_type in NO_INTERP:
        seg.orth = seg.translit
      else:
        seg.orth = xml_morph_seg[0].xpath('./tei:fs/tei:f[@name="orth"]/tei:string', namespaces = namespaces)[0].text
      # przypisz interpretacje - dla elementów nieinterpretowanych pusta interpretacja
      if xml_type in NO_INTERP:
        seg.interps = ''
      else:
        seg.interps = self.make_interps_JSON(seg, xml_morph_seg[0].xpath('./tei:fs/tei:f[@name="interps"]/tei:fs[@type="lex"]', namespaces = namespaces))
    seg.save()
    
  #wycina z oryginalnego text.xml fragment, w którym zapisana jest próbka  
  def get_exc_text(self, dirname, exc_id):
    exc = ''
    exc_found = False
    text_file = io.open(os.path.join(dirname, 'text.xml'), 'r', encoding='utf-8')
    for line in text_file:
      if not exc_found:
        if 'xml:id="txt_{}-div"'.format(exc_id) in line:
          exc_found = True
          exc += line
      else:
        exc += line
        if '</div>' in line:
          break
    text_file.close()
    return exc
          
    
    
  #ładuje nowe fragmenty do bazy z podanego katalogu
  @transaction.atomic
  def __load_excerpts(self, dirname, text_obj):
    corresp = ''

    try:
      old_excerpts = Excerpt.objects.filter(text = text_obj).order_by('xml_id')
      if len(old_excerpts):
        if old_excerpts.last().xml_id != len(old_excerpts):
          raise ValueError(u'Max excerpt xml_id number {} is not total excerpts sum {} for {}'.format(old_excerpts.last().xml_id, len(old_excerpts), dirname))
      exc_start = len(old_excerpts) # najwyższy numer próbki dla danego tekstu istniejący w bazie

      #self.stdout.write(u'Loading excerpts: %s.' % text_obj.text_name)
      morph_tree = etree.parse(os.path.join(dirname, 'ann_morphosyntax.xml'))
      morph_seg_list = morph_tree.xpath('/tei:teiCorpus/tei:TEI/tei:text/tei:body/tei:p/tei:s/tei:seg', namespaces = namespaces)
      morph_index = 0
      seg_tree = etree.parse(os.path.join(dirname, 'ann_segmentation.xml'))
      
      #Zakładam, że ameba pomija znaczniki 'ab', czyli 'p' odpowiada próbce
      xml_excerpts = seg_tree.xpath('/tei:teiCorpus/tei:TEI/tei:text/tei:body/tei:p', namespaces = namespaces)
      for exc_id, xml_p in enumerate(xml_excerpts, start = 1):
        p_id = xml_p.get('{http://www.w3.org/XML/1998/namespace}id')
        if p_id is None:
          raise ValueError(u'Excerpt without id in {}!'.format(os.path.join(dirname, 'ann_segmentation.xml')))
        elif p_id != 'segm_{}-p'.format(exc_id):
          raise ValueError(u'Excerpt number {} has no corresponding id: {}!'.format(exc_id, p_id))
        else:
          if exc_id > exc_start:
            self.stdout.write(u'Excerpt {}'.format(exc_id))

            #s = time.clock()

            #stwórz i zapisz nowy obiekt Excerpt w bazie 
            exc_xml = self.get_exc_text(dirname, exc_id)
            exc = Excerpt(text = text_obj, xml_id = exc_id, xml_content = exc_xml)
            exc.save()
            segs = 1  #numer kolejny segmentu w ramach próbki

            for entry_id, xml_segentry in enumerate(xml_p.xpath('./tei:s/*', namespaces = namespaces)): #w zdaniu 's' może być segment lub choice
              #xml_corresp = xml_seg.get('corresp').split('.')[1] #na indeksie [0] jest 'text' z napisu 'text.xml'
              
              tag = self.tag_wo_ns(xml_segentry)
              #self.stdout.write(u'Entry_id: {}, tag: {}'.format(entry_id + 1, tag))

              if tag == 'seg':
                #self.stdout.write(u'Segment: {}, morph_index: {}'.format(segs, morph_index))
                i, morph_seg = self.find_morph(xml_segentry.get('{http://www.w3.org/XML/1998/namespace}id'), morph_index, morph_seg_list)
                self.make_segment(exc, entry_id + 1, xml_segentry, segs, -1, -1, morph_seg) # segment bezpośrednio w 's'
                morph_index = i
                segs += 1
              elif tag == 'choice':
                for choice_id, xml_choice in enumerate(xml_segentry):
                  tag_choice = self.tag_wo_ns(xml_choice)
                  if tag_choice == 'seg':
                    #self.stdout.write(u'Segment: {}, morph_index: {}'.format(segs, morph_index))
                    i, morph_seg = self.find_morph(xml_choice.get('{http://www.w3.org/XML/1998/namespace}id'), morph_index, morph_seg_list)
                    self.make_segment(exc, entry_id + 1, xml_choice, segs, choice_id + 1, -1, morph_seg) # segment bezpośrednio w 'choice' - wszystkie takie dostają ten sam entry_id
                    morph_index = i
                    segs += 1
                  elif tag_choice == 'paren':
                    for paren_id, xml_paren in enumerate(xml_choice):
                      tag_paren = self.tag_wo_ns(xml_paren)
                      if tag_paren == 'seg':
                        #self.stdout.write(u'Segment: {}, morph_index: {}'.format(segs, morph_index))
                        i, morph_seg = self.find_morph(xml_paren.get('{http://www.w3.org/XML/1998/namespace}id'), morph_index, morph_seg_list)
                        self.make_segment(exc, entry_id + 1, xml_paren, segs, choice_id + 1, paren_id + 1, morph_seg) # segment w 'paren'
                        morph_index = i
                        segs += 1
                      else:
                        raise ValueError(u'{} w paren, tekst: {}, nr segmentu: {}'.format(tag_paren, text_obj.text_name, segs))
                  else:
                    raise ValueError(u'{} w choice, tekst: {}, nr segmentu: {}'.format(tag_choice, text_obj.text_name, segs))
              else:
                raise ValueError(u'{} w zdaniu, tekst: {}, nr segmentu: {}'.format(tag, text_obj.text_name, segs))

            #self.stdout.write("czas ladowania: {}".format(time.clock()-s))  
    
          
    except etree.LxmlError:
      fail_str = u'Failed to read xml tree for %s' % text_obj.text_name
      self.stdout.write('{}'.format(fail_str))
      raise
    except ValueError as err:
      fail_str = err.args
      self.stdout.write('{}'.format(fail_str))
      raise
    except:
      fail_str = u'Failed to process %s, unexpected error.' % text_obj.text_name
      self.stdout.write('{}'.format(fail_str))
      raise
