# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.db.models import Max
from annotation.models import Text, Excerpt, Segment, Workdetail
import os, io, errno
from lxml import etree
from itertools import product
import json
import time
from annotation.anot2_config import REF_MODE, NO_INTERP
from shutil import rmtree

namespaces = {
  'tei': 'http://www.tei-c.org/ns/1.0',
  'xi': 'http://www.w3.org/2001/XInclude',
  'nkjp': 'http://www.nkjp.pl/ns/1.0'
}

def write_header(f, t = None):
    f.write(u'''<?xml version="1.0" encoding="UTF-8"?>
<teiCorpus xmlns:xi="http://www.w3.org/2001/XInclude" xmlns:nkjp="http://www.nkjp.pl/ns/1.0" xmlns="http://www.tei-c.org/ns/1.0">
  <xi:include href="KORBA_header.xml"/>
  <TEI>
    <xi:include href="header.xml"/>
''')
    if t:
        f.write(u'''    <text xml:id="''' + t + '''_text" xml:lang="pl">
      <body xml:id="''' + t + '''_body">
''')
    else:
        f.write(u'''    <text>
      <body>
''')


def write_footer(f):
    f.write(u'''      </body>
    </text>
  </TEI>
</teiCorpus>
''')

'''
Wypisanie do pliku "f" tagu rozpoczynającego "tg" z wcięciem na "indent"-krotną podwójną spację.
Jeśli "newl" = True, to za tagiem otwierającym wpisywany jest koniec linii.
Jeśli "emptytg" = True, to znacznik traktowany jest jako pusty i jednocześnie kończony.
Na "**attribs" przekazywane są atrybuty do wypisania.
'''
def start_tag(tg, indent, newl, emptytg, xml_id, **attribs):
    line = "  "*indent + u"<{}".format(tg)
    if xml_id != "":
      line += u" xml:id=\"{}\"".format(xml_id)
    for att_name in attribs.keys():
        line += u" {}=\"{}\"".format(att_name, attribs[att_name])
    if emptytg:
      line += u"/>"
    else:
      line += u">"
    if newl:
      line += "\n"
    return line

def end_tag(tg, indent):
    return "  "*indent + u"</{}>\n".format(tg)

# Robi z zapisu z kropką listę zapisów pojedynczych
def unpack_msd(msd):
    msd_list = []
    parts = [part.split('.') for part in msd.split(':')]
    for p in product(*parts):
        msd_list.append(':'.join(p))
    return msd_list

'''Wypisywanie informacji o zakończonych próbkach do plików
ann_segmentation.xml oraz ann_morphosyntax.xml
'''

class Command(BaseCommand):
  help = 'Writes completed excerpts from Anotatornia2 database in TEI format.'

  def add_arguments(self, parser):
    # Named (optional) arguments
    parser.add_argument('--nochoice',
      action='store_true',
      dest='nochoice',
      default=False,
      help='Exports only choosen branch of choice as plain segments')
    parser.add_argument('--dirname',
      dest='dirname',
      nargs = '?',
      default='',
      help='Directory name for output (will be erased if exists)')

  def handle(self, *args, **options):
    self.stdout.write('Exporting excerpts from Anotatornia2...')

    if options['dirname']:
      outdir = options['dirname']
    else:
      lt = time.localtime(time.time())
      outdir = u'anot2_{}{:02d}{:02d}{:02d}{:02d}{:02d}'.format(lt.tm_year, lt.tm_mon, lt.tm_mday, lt.tm_hour, lt.tm_min, lt.tm_sec)
    
    try:
      os.mkdir(outdir)
    except OSError as exc:
      if exc.errno != errno.EEXIST:
        self.stdout.write(u'Failed to make directory {}.'.format(outdir))
        raise
      else: # jeśli folder istniał, to usuń dotychczasową zawartość
        for sd in os.listdir(outdir):
          rmtree(os.path.join(outdir, sd))

    self.__write(outdir, options['nochoice'])
    self.stdout.write('Results written to {}'.format(outdir))

    
  #eksportuje gotowe próbki do katalogu texts_dir
  def __write(self, texts_dir, nochoice):
    #weź listę identyfikatorów tekstów, w których są skończone próbki
    text_list = Text.objects.filter(excerpt__details__work_status__gt=5).values_list('id', flat=True).order_by('id').distinct()

    for text_id in text_list:
        text_name = Text.objects.get(id = text_id).text_name
        self.stdout.write(text_name)
        dirname = os.path.join(unicode(texts_dir), text_name)
        
        #Stwórz folder z nazwą tekstu
        try:
          os.mkdir(dirname)
        except:
          self.stdout.write(u'Failed to make directory {}.'.format(dirname))
          raise
        
        #Stwórz pliki ann_morphosyntax.xml oraz ann_segmentation.xml
        try:
          text_file = io.open(os.path.join(dirname, 'text.xml'), "w", encoding='utf-8')
          header_file = io.open(os.path.join(dirname, 'header.xml'), "w", encoding='utf-8')
          segm_file = io.open(os.path.join(dirname, 'ann_segmentation.xml'), "w", encoding='utf-8')
          morph_file = io.open(os.path.join(dirname, 'ann_morphosyntax.xml'), "w", encoding='utf-8')
        except:
          self.stdout.write(u'Failed to create file in {}/{} for output'.format(texts_dir, text_name))
          continue

        # Zapisz plik z nagłówkiem tekstu
        header_file.write(Text.objects.get(id = text_id).header)
        header_file.close()
        
        # Znajdź próbki z danego tekstu
        exc_list = Excerpt.objects.filter(text = text_id, details__work_status__gt = 5).order_by('xml_id').distinct()
        
        if len(exc_list) > 0:
          write_header(text_file)
          write_header(segm_file, 'segm') 
          write_header(morph_file, 'morph')
          
          for exc in exc_list:
            self.stdout.write(u'excerpt: {}'.format(exc.xml_id))
            
            text_file.write(exc.xml_content)
            
            if REF_MODE:
              # Wykluczamy ukryte segmenty - informacja nie pojawi się ani w ann_segmentation ani ann_morphosyntax
              wd = exc.details.all()[0]
              # Sprawdź, czy próbka była z superanotacją, czy bez.
              if wd.work_status == 6:
                if nochoice:
                  seg_list = Segment.objects.filter(excerpt = exc, user = wd.user).exclude(is_hidden = True).exclude(is_rejected = -1).order_by('xml_id', 'choice_id', 'paren_id')
                else:
                  seg_list = Segment.objects.filter(excerpt = exc, user = wd.user).exclude(is_hidden = True).order_by('xml_id', 'choice_id', 'paren_id')
              else:
                if nochoice:
                  seg_list = Segment.objects.filter(excerpt = exc, ref_rejected = 1).exclude(ref_id = 0).exclude(is_hidden = True).exclude(is_rejected = -1).order_by('xml_id', 'choice_id', 'paren_id')
                else:
                  seg_list = Segment.objects.filter(excerpt = exc, ref_rejected = 1).exclude(ref_id = 0).exclude(is_hidden = True).order_by('xml_id', 'choice_id', 'paren_id')

            else:  # not in REF_MODE
              if nochoice:
                seg_list = Segment.objects.filter(excerpt = exc).exclude(user_id__isnull = True).exclude(is_hidden = True).exclude(is_rejected = -1).order_by('xml_id', 'choice_id', 'paren_id')
              else:
                seg_list = Segment.objects.filter(excerpt = exc).exclude(user_id__isnull = True).exclude(is_hidden = True).order_by('xml_id', 'choice_id', 'paren_id')
              

            seg_len = len(seg_list)
            i = 0
            if seg_len > 0:
              segm_file.write(start_tag('p', 4, True, False, 'segm_{}-p'.format(exc.xml_id), corresp='text.xml#txt_{}-div'.format(exc.xml_id)))
              morph_file.write(start_tag('p', 4, True, False, 'morph_{}-p'.format(exc.xml_id), corresp='ann_segmentation.xml#segm_{}-p'.format(exc.xml_id)))
              segm_sent = ''
              morph_sent = ''
              seg_indent = 6
              in_choice = False
              in_paren = False
              in_choice_id = 0
              in_paren_id = 0

              for seg_id, seg in enumerate(seg_list, start=1):
                #Zbierz zdanie w ann_segmentation
                
                if not(nochoice):
                  if seg.choice_id is None:
                    choice_id = 0
                  else:
                    choice_id = seg.choice_id
                    
                  if seg.paren_id is None:
                    paren_id = 0
                  else:
                    paren_id = seg.paren_id

                  if in_paren and (paren_id < in_paren_id):
                    seg_indent -= 1
                    segm_sent += end_tag('nkjp:paren', seg_indent)
                    in_paren = False

                  if in_choice and (choice_id < in_choice_id):
                    seg_indent -= 1
                    segm_sent += end_tag('choice', seg_indent)
                    in_choice = False
                    
                  if (not in_choice) and (choice_id == 1):
                     in_choice = True
                     segm_sent += start_tag('choice', seg_indent, True, False, '')
                     seg_indent += 1

                  if (not in_paren) and (paren_id == 1):
                     in_paren = True
                     segm_sent += start_tag('nkjp:paren', seg_indent, True, False, '')
                     seg_indent += 1
                  
                  in_choice_id = choice_id
                  in_paren_id = paren_id
                
                attribs = {}
                if seg.corresp_pos is not None:
                  attribs['corresp'] = 'text.xml#string-range({},{},{})'.format(seg.corresp_elem, seg.corresp_pos, seg.corresp_len)
                if seg.seg_type != '':
                  attribs['type'] = seg.seg_type
                if seg.nps:
                  attribs['nkjp:nps'] = 'true'
                if (seg.is_rejected == -1):
                  attribs['nkjp:rejected'] = 'true'
                if not seg.korbeusz:
                  attribs['nkjp:korbeusz'] = 'false'
                if seg.translit == '':
                  segm_sent += start_tag('seg', seg_indent, True, True, 'segm_{}.{}-seg'.format(exc.xml_id, seg_id), **attribs)
                else:
                  segm_sent += start_tag('seg', seg_indent, True, False, 'segm_{}.{}-seg'.format(exc.xml_id, seg_id), **attribs)
                  segm_sent += start_tag('w', seg_indent + 1, False, False, '') + seg.translit + end_tag('w', 0)
                  segm_sent += end_tag('seg', seg_indent) 
                
                #Zbierz zdanie w ann_morphosyntax
#                if not((seg.is_rejected == -1) or seg.is_hidden or (seg.interps == "")): 
                if (seg.is_rejected != -1) and (seg.interps or (seg.seg_type in NO_INTERP)): # Dorzucamy do ann_morphosyntax opis dla segmentów nieinterpretowanych z pustą interpretacją
                  morph_sent += start_tag('seg', 6, True, False, 'morph_{}.{}-seg'.format(exc.xml_id, seg_id), corresp='ann_segmentation.xml#segm_{}.{}-seg'.format(exc.xml_id, seg_id))
                  morph_sent += start_tag('fs', 7, True, False, '', **{'type': 'morph'})
                  morph_sent += start_tag('f', 8, True, False, '', name = "orth")
                  morph_sent += '  '*9 + u'<string>' + seg.orth + '</string>\n'
                  morph_sent += end_tag('f', 8)
                  morph_sent += start_tag('f', 8, True, False, '', name = "translit")
                  morph_sent += '  '*9 + u'<string>' + seg.translit + '</string>\n'
                  morph_sent += end_tag('f', 8)
                  if seg.nps:
                    morph_sent += start_tag('f', 8, True, False, '', name = "nps")
                    morph_sent += '  '*9 + u'<binary value="true"/>\n'
                    morph_sent += end_tag('f', 8)
                  morph_sent += start_tag('f', 8, True, False, '', name = "interps")

                  msd_choosen = []
                  if seg.interps:
                    all_interps = json.loads(seg.interps)
                    interp_choosen = seg.radio_id.split('.')
                    for interp_id, interp in enumerate(all_interps, start=1):
                      if (interp_id == int(interp_choosen[0])+1) and ((interp['base'] != seg.base) or (interp['ctag'] != seg.ctag)):
                        print u"Błąd wersji w segmencie o id={}: {}!={}, {}!={}".format(seg.id, interp['base'], seg.base, interp['ctag'], seg.ctag).encode('utf-8')
                      morph_sent += start_tag('fs', 9, True, False, 'morph_{}.{}.{}-lex'.format(exc.xml_id, seg_id, interp_id), **{'type': 'lex'})
                      morph_sent += start_tag('f', 10, True, False, '', name = "base")
                      morph_sent += '  '*11 + u"<string>" + interp['base'] + "</string>\n"
                      morph_sent += end_tag('f', 10)
                      morph_sent += start_tag('f', 10, True, False, '', name = "ctag")
                      morph_sent += '  '*11 + u"<symbol value=\"" + interp['ctag'] + "\"/>\n"
                      morph_sent += end_tag('f', 10)
                      morph_sent += start_tag('f', 10, True, False, '', name = "msd")
                      msd_list = interp['msd']
                      valt_end = False
                      if (len(msd_list) == 1) and ('.' not in msd_list[0]['val']):
                        msd_indent = 11
                      else:
                        msd_indent = 12
                        morph_sent += '  '*11 + u"<vAlt>\n"
                        valt_end = True
                      i = 1
                      for msd_id, msd in enumerate(msd_list, start=1):
                        if ('.' in msd['val']):
                          unpacked = unpack_msd(msd['val'])
                        else:
                          unpacked = [msd['val']]
                        if (interp_id == int(interp_choosen[0])+1) and (msd_id == int(interp_choosen[1])+1):
                          msd_choosen = unpacked
                        for val in unpacked:
                          attribs = {'value': val}
                          if msd['manual']:
                            attribs['nkjp:manual'] = "true"
                          morph_sent += start_tag('symbol', msd_indent, True, True, "morph_{}.{}.{}.{}-msd".format(exc.xml_id, seg_id, interp_id, i), **attribs)
                          i += 1
                      if valt_end:
                        morph_sent += '  '*11 + u"</vAlt>\n"
                      morph_sent += end_tag('f', 10)
                      morph_sent += end_tag('fs', 9)
                  
                  elif (seg.seg_type in NO_INTERP): # wygenerowanie interpretacji z pustymi wartościami dla segmentów nieinterpretowanych
                    morph_sent += start_tag('fs', 9, True, False, 'morph_{}.{}.1-lex'.format(exc.xml_id, seg_id), **{'type': 'lex'})
                    morph_sent += start_tag('f', 10, True, False, '', name = "base")
                    morph_sent += '  '*11 + u"<string></string>\n"
                    morph_sent += end_tag('f', 10)
                    morph_sent += start_tag('f', 10, True, False, '', name = "ctag")
                    morph_sent += '  '*11 + u"<symbol value=\"\"/>\n"
                    morph_sent += end_tag('f', 10)
                    morph_sent += start_tag('f', 10, True, False, '', name = "msd")
                    morph_sent += start_tag('symbol', 11, True, True, "morph_{}.{}.1.1-msd".format(exc.xml_id, seg_id), **{'value': ''})
                    morph_sent += end_tag('f', 10)
                    morph_sent += end_tag('fs', 9)

                  morph_sent += end_tag('f', 8)

                  # Jeśli segment nie został odrzucony w segmentacji - is_rejected != -1 - to wypisz ujednoznacznienie segmentu
                  if ((seg.is_rejected != -1) and seg.interps):
                    s = u"{}:{}".format(seg.base, seg.ctag)
                    morph_sent += start_tag('f', 8, True, False, '', name = "disamb")
                    i = 1
                    for val in msd_choosen:
                      s1 = s + u":" + val if val else s
                      morph_sent += start_tag('fs', 9, True, False, '', **{'type': 'tool_report'})
                      morph_sent += start_tag('f', 10, True, True, '', fVal = "#morph_{}.{}.{}.{}-msd".format(exc.xml_id, seg_id, int(interp_choosen[0])+1, int(interp_choosen[1])+i), name = "choice")
                      morph_sent += start_tag('f', 10, True, False, '', name = "interpretation")
                      morph_sent += '  '*11 + u"<string>" + s1 + "</string>\n"
                      morph_sent += end_tag('f', 10)
                      morph_sent += end_tag('fs', 9)                  
                      i += 1
                    morph_sent += end_tag('f', 8)
                    
                  morph_sent += end_tag('fs', 7)
                  morph_sent += end_tag('seg', 6)
                
                #Zapisz zdania do plików
                if seg.sent_end:
                  segm_sent = start_tag('s', 5, True, False, 'segm_{}.{}-s'.format(exc.xml_id, seg_id)) + segm_sent + end_tag('s', 5)
                  segm_file.write(segm_sent)
                  segm_sent = ''
                  morph_sent = start_tag('s', 5, True, False, 'morph_{}.{}-s'.format(exc.xml_id, seg_id), corresp='ann_segmentation.xml#segm_{}.{}-s'.format(exc.xml_id, seg_id)) + morph_sent + end_tag('s', 5)
                  morph_file.write(morph_sent)
                  morph_sent = ''
              segm_file.write(end_tag('p', 4))
              morph_file.write(end_tag('p', 4))
            else:
              self.stdout.write(u'No segments for {} excerpt in {}'.format(exc.xml_id, text_name))
          
          write_footer(text_file)
          write_footer(segm_file)
          write_footer(morph_file)
        else:
          self.stdout.write(u'No excerpts in {}'.format(text_name))
        
        segm_file.close()
        morph_file.close()



